/*******************************************************************************
* NXP Semiconductors
* (c) Copyright 2016 NXP Semiconductors
* ALL RIGHTS RESERVED.
********************************************************************************
Services performed by NXP in this matter are performed AS IS and without any 
warranty. CUSTOMER retains the final decision relative to the total design
and functionality of the end product. NXP neither guarantees nor will be held
liable by CUSTOMER for the success of this project.
NXP DISCLAIMS ALL WARRANTIES, EXPRESSED, IMPLIED OR STATUTORY INCLUDING,
BUT NOT LIMITED TO, IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS FOR
A PARTICULAR PURPOSE ON ANY HARDWARE, SOFTWARE ORE ADVISE SUPPLIED 
TO THE PROJECT BY NXP, AND OR NAY PRODUCT RESULTING FROM NXP SERVICES. 
IN NO EVENT SHALL NXP BE LIABLE FOR INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF THIS AGREEMENT. 
CUSTOMER agrees to hold NXP harmless against any and all claims demands 
or actions by anyone on account of any damage, or injury, whether commercial,
contractual, or tortuous, rising directly or indirectly as a result 
of the advise or assistance supplied CUSTOMER in connection with product, 
services or goods supplied under this Agreement.
********************************************************************************
* File:             main.c
* Owner:            David Tosenovjan
* Version:          0.0
* Date:             Apr-11-2017
* Classification:   General Business Information
* Brief:            Brief description of the file
********************************************************************************
* Detailed Description:
* Example shows MCU's temperature measurement using internal temperature sensor.
* After basic initialization, the code initializes and calibrates eQADC, reads 
* test flash content and performs all necessary ADC measurement to calculate die 
* temperature. Results are then displayed in the terminal window. 
* ------------------------------------------------------------------------------
* Test HW:         XPC567XKIT516 - MPC5674ADAT516 Rev.C, MPC567XEVBFXMB Rev.B
* MCU:             PPC5674FMVYA264
* Terminal:        19200-8-no parity-1 stop bit-no flow control on eSCI_A
* Fsys:            264/200/150/60 MHz
* Debugger:        Lauterbach Trace32
*                  PeMicro USB-ML-PPCNEXUS
* Target:          RAM, internal_FLASH
* EVB connection:  default
*
********************************************************************************
Revision History:
Ver  Date         Author            Description of Changes
0.0  Apr-11-2017  David Tosenovjan  Initial version
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h> /* qsort */
#include "MPC5674F_MVxA264.h"
#include "Exceptions.h"
#include "IntcInterrupts.h"
#include "uart.h"
#include "clock.h"
#include "Optimizations.h"
#include "eQADC_macros_MAMBA.h"

/*******************************************************************************
* Global variables
*******************************************************************************/
uint32_t TSENS_CODE_T_LOW;
uint32_t TSENS_CODE_T_HIGH;
uint32_t VBG_CODE_T_LOW;
int32_t T_LOW, T_HIGH;
    
uint32_t VBG_T_Ch144 = 0;
uint32_t VBG_T_Ch45 = 0;
uint32_t VTSENS_T = 0;

uint32_t VRH, VRL, Vref, VBG_CODE_T, TSENS_CODE_T;

double beta;

double SoC_Temp = 1.00;
int32_t SoC_Temp_int = 0;


/*******************************************************************************
* Constants and macros
*******************************************************************************/

#define MPC567xADAT416 1
#define MPC567xADAT516 2

/* Choose mini-module plugged in the motherboard (one of 2 options above) */
#define USED_BOARD MPC567xADAT516

#if USED_BOARD==MPC567xADAT516
    #define INITIALIZE_EXTERNAL_SRAM 1
#endif

/* Enable caches, branch prediction and optimize flash setting */
#define OPTIMIZATIONS_ON 0

#define IDEAL_RES75 0x3000		// Ideal value of 75%(VDH-VDL)
#define IDEAL_RES25 0x1000		// Ideal value of 25%(VDH-VDL)

/* EQADC_CAL = 1 means eQADC conversion with Calibration */
/* EQADC_CAL = 0 means eQADC conversion uncalibrated */
#define EQADC_CAL 1


/*******************************************************************************
* Function prototypes
*******************************************************************************/
static void FMPLL_init(void);
static void DisableWatchdog(void);
static void eQADCA_init(void);

static uint32_t ConvertChannel(int32_t channel, int32_t bCal);
int Compare (const void * p0, const void * p1);
static uint32_t ConvertChannel_Median(int32_t channel, int32_t bCal);
static int32_t EQADC_A_PowerUpCalibration(void);

static void Read_TSENS_constants_from_test_flash(void);
static void Determine_SoC_temperature(void);
                   

/*******************************************************************************
* Local functions
*******************************************************************************/

/*******************************************************************************
Function Name : HW_init
Engineer      : David Tosenovjan
Date          : Nov-22-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : initialization of the hw for the purposes of this example
Issues        : 
*******************************************************************************/
static void HW_init(void)
{
    DisableWatchdog();
    FMPLL_init();
    
    /* set round robin for all slaves */
    XBAR.SGPCR0.B.ARB = 1;
    XBAR.SGPCR1.B.ARB = 1;
    XBAR.SGPCR2.B.ARB = 1;
    XBAR.SGPCR6.B.ARB = 1;
    XBAR.SGPCR7.B.ARB = 1;
    
    #if OPTIMIZATIONS_ON
        Optimizations();
    #endif    

}


/*******************************************************************************
Function Name : DisableWatchdog
Engineer      : David Tosenovjan
Date          : Mar-14-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : disable SWT and e200 Core Watchdog Timer
Issues        : 
*******************************************************************************/
static void DisableWatchdog(void)
{    
    /*
    REM Disable Watchdog Timers
    REM MCM SWT New SWT in MPC56xx devices
    mm.l $fff38000 $ff00000A
    REM e200 Core Watchdog Timer (all MPC55xx and MPC56xx devices)
    spr 340t 0
    */
    SWT.MCR.R = 0xFF00000A;
    
    asm
    {
        mfhid0  r11		  	  /* Move from spr HID0 to r11 (copies HID0) */
        li      r12, 0xBFFF  /* Load immed. data of ~0x4000 to r12 */
        oris    r12, r12, 0xFFFF
        and     r11, r12, r11  /* OR r12 (~0x00004000) with r11 (HID0 value) */
        mthid0  r11          /* Move result to HID0 */

        lis     r12, 0x00
        mtspr   tcr, r12  /* clear SPR TCR register */
    }
}


/*******************************************************************************
Function Name : FMPLL_init
Engineer      : David Tosenovjan
Date          : Apr-19-2012
Parameters    : 
Modifies      : 
Returns       : 
Notes         : Initializes PLL for 264/200/150/60MHz system clock                         
Issues        : - Assuming 40MHz crystal reference
                - clock.h must be included
*******************************************************************************/
static void FMPLL_init(void)
{
    #if defined(CORE_CLOCK_264MHz)
        FMPLL.ESYNCR2.R = 0x00000002;								
        FMPLL.ESYNCR1.R = 0x70040032;								
        while (FMPLL.SYNSR.B.LOCK != 1) {}; /* Wait for FMPLL to LOCK  */								
        FMPLL.ESYNCR2.R = 0x00000001;                  /* Fsys =264Mhz */								
        printf("fsys = 264MHz\n\r");
    
    #elif defined(CORE_CLOCK_200MHz)
        FMPLL.ESYNCR2.R = 0x00000002;								
        FMPLL.ESYNCR1.R = 0x70040022;								
        while (FMPLL.SYNSR.B.LOCK != 1) {}; /* Wait for FMPLL to LOCK  */								
        FMPLL.ESYNCR2.R = 0x00000001;                  /* Fsys =200Mhz */								
        printf("fsys = 200MHz\n\r");
    
    #elif defined(CORE_CLOCK_150MHz)
        FMPLL.ESYNCR2.R = 0x00000005;								
        FMPLL.ESYNCR1.R = 0x7003002C;								
        while (FMPLL.SYNSR.B.LOCK != 1) {}; /* Wait for FMPLL to LOCK  */								
        FMPLL.ESYNCR2.R = 0x00000003;                  /* Fsys =150Mhz */								
        printf("fsys = 150MHz\n\r");
    #else
        printf("FMPLL wasn't set properly - default 60MHz fsys used\n\r");
    #endif
}


/*******************************************************************************
Function Name : eQADC_A_init
Engineer      : David Tosenovjan
Date          : Apr-11-2017
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void eQADCA_init(void)
{

    // Send CFIFO0 a ADC0/1 configuration commands 
    // enable ADC0/1 & sets prescaler= divide by 10 
    
    PUSH_TO_CFIFO0_A ( B0                     |
                       ADC_REG_VALUE (0x8004) | // enables ADC0
                       ADC_REG_ADDR (ADC0_CR) );
                                 
    PUSH_TO_CFIFO0_A ( EOQ                    |
                       B1                     |
                       ADC_REG_VALUE (0x8004) | // enables ADC1
                       ADC_REG_ADDR (ADC1_CR) );                             
                                      
    // Trigger CFIFO 0 using Single Scan SW mode                                  
    EQADC_A.CFCR[0].R = 0x0410;
    
    
    // Wait for End Of Queue flag   
    while (EQADC_A.FISR[0].B.EOQF !=1)
    {
    } 
    
    // Clear End Of Queue flag 
    EQADC_A.FISR[0].R = 0x10000000;  
}



/*******************************************************************************
Function Name : ConvertChannel
Engineer      : David Tosenovjan
Date          : Apr-11-2017
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static uint32_t ConvertChannel(int32_t channel, int32_t bCal)
{
    uint32_t ret = 0;

    /* Conversion command: convert "channel" */
    /* with ADC0, set EOQ, and send result to RFIFO 0 */ 
    /* 128 adc clocks sampling */
    EQADC_A.CFPR[0].R = 0x800C0000 | (channel<<8) | (bCal<<24); 
    EQADC_A.CFCR[0].R = 0x0410;   /* Trigger CFIFO 0 using Single Scan SW mode */
    /* Wait for RFIFO 0's Drain Flag to set */
    while (EQADC_A.FISR[0].B.RFDF != 1){}
    
    ret = EQADC_A.RFPR[0].R;
    
    /* Clear RFIFO 0's Drain Flag */
    /* Clear CFIFO's End of Queue flag */
    /* RFDF + EOQF */
    EQADC_A.FISR[0].R = 0x10020000;
    
    return ret;
}

/*******************************************************************************
Function Name : Compare
Engineer      : b08110
Date          : Sep-02-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
int Compare (const void * p0, const void * p1)
{
    uint32_t a0 = (uint32_t) *(uint32_t*)p0;
    uint32_t a1 = (uint32_t) *(uint32_t*)p1;

    if(a0==a1) return 0;
    if(a0<a1) return -1;
    else return 1;
}

/*******************************************************************************
Function Name : ConvertChannel_Median
Engineer      : b08110
Date          : Sep-02-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : Performs 21 conversions and choose the value in the midlle
Issues        : 
*******************************************************************************/
static uint32_t ConvertChannel_Median(int32_t channel, int32_t bCal)
{
    uint32_t results[21];
            
    int32_t i = 0;
    for(i=0;i<sizeof(results)/sizeof(results[0]);i++)
    {
        results[i] = ConvertChannel(channel, bCal);
    }
    
    /* sort results */
    qsort((void*)results,11,sizeof(results[i]),Compare);
    
    return results[10]; /* median */
}


/*******************************************************************************
Function Name : EQADC_A_PowerUpCalibration
Engineer      : David Tosenovjan
Date          : Apr-11-2017
Parameters    : 
Modifies      : 
Returns       : 
Notes         : Based on recommnedation in AN2989
Issues        : 75% and 25%VRH (43 and 44) must be converted without calibration
*******************************************************************************/
static int32_t EQADC_A_PowerUpCalibration(void) 
{
    unsigned long raw_res75, raw_res25;
    unsigned long cal_res75, cal_res25;
    
    float gccf, occf;
    
    unsigned short gain;
    short offset;
    
    /* Convert channel 43 (75%VRH)with CAL=0 */
    raw_res75 = ConvertChannel_Median(43,0);
    /* Convert channel 44 (25%VRH)with CAL=0, 64 sample time (errata) */
    raw_res25 = ConvertChannel_Median(44,0);
   
    /* Compute Gain and Offset */
    gccf = (float)(8192.0)/(raw_res75 - raw_res25);
    gain = gccf * 16384;
    occf = (float)(IDEAL_RES75 - (gccf * raw_res75) - 2.0);
    offset = occf;

    /* now store Gain in ADCn_GCCR */
    EQADC_A.CFPR[0].R = (gain<<8) | 0x00000004;
    /* now store offset in ADCn_OCCR */
    EQADC_A.CFPR[0].R = ((unsigned short)offset << 8) | 0x00000005;

    /* now store Gain in ADCn_AGR */
    EQADC_A.CFPR[0].R = (gain<<8) | 0x00000031;
    /* now store offset in ADCn_ACR */
    EQADC_A.CFPR[0].R = ((unsigned short)offset << 8) | 0x00000032 | 0x80000000;
    
    EQADC_A.CFCR[0].R = 0x0410;    /* Trigger CFIFO 0 using Single Scan SW mode */
    while (EQADC_A.FISR[0].B.EOQF !=1) {} /* Wait for End Of Queue flag */
    EQADC_A.FISR[0].R = 0x10000000;
    
    cal_res75 = ConvertChannel(43,1);
    cal_res25 = ConvertChannel(44,1);    
    
    // Check the results
    if ((abs(cal_res75 - IDEAL_RES75) > 0xF) || (abs(cal_res25 - IDEAL_RES25) > 0xF)) 
    {
	    /* error */
	    return 1;
    }
    
    return 0; /* no error */    
}

/*******************************************************************************
Function Name : Determine_SoC_temperature
Engineer      : Apr-11-2017 
Date          : David Tosenovjan
Parameters    : 
Modifies      : 
Returns       : 
Notes         : Measure on-chip temperature sensor and compute die temperature
Issues        : MPC5634M = 144, MPC5674F = 45 (bandgap channel)
*******************************************************************************/
static void Determine_SoC_temperature(void)
{
    
    VRH = ConvertChannel(40,EQADC_CAL);
    VRL = ConvertChannel(41,EQADC_CAL);
    
    Vref = VRH - VRL;
    
    VBG_T_Ch45 = ConvertChannel(45,EQADC_CAL);
    
    VBG_CODE_T = (uint32_t)((double)(VBG_T_Ch45)/Vref*(1<<14));
    
    VTSENS_T = ConvertChannel(128,EQADC_CAL);
    TSENS_CODE_T = (uint32_t)((double)(VTSENS_T)/Vref*(1<<14));

    // for old MCU's datecodes VBG_CODE_T_LOW may not be stored in test flash
    // in such case set beta for 1
    if (VBG_CODE_T_LOW != 0xFFFF)   
        beta = (double)VBG_CODE_T_LOW/(double)VBG_CODE_T;
    else 
        beta = 1.0;
       
    //             (ADC_reading * B)-(VTSEN(TLow))
    // T = TLow + --------------------------------- * (THigh - TLow)
    //              (VTSEN(THigh)) - (VTSEN(TLow))     
    //
    // B = Vbandgap(Tlow)/Vbandgap
    //  VBG(TLow)    : bandgap voltage at low calibration temperature (TLow)
    //  VTSENS(TLow) : temp sensor voltage at low calibration temperature (TLow)
    //  VTSENS(THigh): temp sensor voltage at high calibration temperature (THigh)
    
    SoC_Temp = (double)T_LOW;
    
    SoC_Temp += (TSENS_CODE_T*beta - TSENS_CODE_T_LOW)*(T_HIGH - T_LOW)/(TSENS_CODE_T_HIGH - TSENS_CODE_T_LOW);
    
    SoC_Temp_int = (int32_t)SoC_Temp;
          
}


/*******************************************************************************
Function Name : Read_TSENS_constants_from_test_flash
Engineer      : David Tosenovjan
Date          : Apr-11-2017 
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void Read_TSENS_constants_from_test_flash(void)
{
    // Read TSENS calibration constants from Test Flash
    TSENS_CODE_T_LOW = TSENS.TCCR0.B.TSCV1;
    TSENS_CODE_T_HIGH = TSENS.TCCR0.B.TSCV2;
    VBG_CODE_T_LOW = TSENS.TCCR1.B.TSCV3;

    // Determine hot factory calibration temperature
    if(1==TSENS.TCCR0.B.THIGH) T_HIGH = 125;
	if(2==TSENS.TCCR0.B.THIGH) T_HIGH = 145;
	if(3==TSENS.TCCR0.B.THIGH) T_HIGH = 150;
    
    // Determine low factory calibration temperature 
    if(0==TSENS.TCCR0.B.TLOW) T_LOW = 25;
    if(1==TSENS.TCCR0.B.TLOW) T_LOW = 40;
    if(2==TSENS.TCCR0.B.TLOW) T_LOW = -45;
}


/*******************************************************************************
* Global functions
*******************************************************************************/

/*******************************************************************************
Function Name : main
Engineer      : David Tosenovjan
Date          : Apr-11-2017 
Parameters    : 
Modifies      : 
Returns       : 
Notes         :                             
Issues        : 
*******************************************************************************/
int main(void) 
{
    int32_t i = 0;
    vint32_t counter = 0;
    uint32_t res = 0;

    /* Hardware initialization */
    HW_init();
    
    // eQADC initialization and calibration
    eQADCA_init();  	
  
	for(i=0;i<100000;i++) {}
    
    // note b05111: - this notation calls the function until it returns 0
    i=1;
    res=0;
    for(;i==1;i=EQADC_A_PowerUpCalibration())
    {     
        res++;   
    }
    
    printf("eQADC module initialized and calibrated\n\r");

    printf("TSENS temperature calculation\n\r");
    printf("\n\r");
    
    Read_TSENS_constants_from_test_flash();
    
                      
     /* Loop forever */
    while (1) 
    {      
        /* delay */
        while(counter < 10000000)
        {
            counter++;
        }  
        counter = 0;
        
        // Convert internal ADC channels and calculate temperature
        Determine_SoC_temperature();
        	
        // Display measured temperature value
        printf("VBG_CODE_T (ch45)  = %d => beta = %.5f\n\r", VBG_CODE_T, beta );
	    printf("TSENS_CODE_T (ch128) = %d\n\r", TSENS_CODE_T );
	    printf("\n\r");
	    printf("Temp = %.2f degC\r", SoC_Temp);
	    printf("\x1b[3A");	// return 3 lines back
          
    }

}



